import QuantLib as ql
import numpy as np
from scipy.optimize import differential_evolution
from typing import List, Tuple
from exotx.data.marketdata import MarketData


class HestonModel:
    """Class for the Heston model."""
    def __init__(self,
                 reference_date: ql.Date,
                 calendar: ql.Calendar,
                 market_data: MarketData) -> None:
        self.reference_date = reference_date
        self.calendar = calendar
        self.market_data = market_data
        self.initial_conditions = (0.02, 0.2, 0.5, 0.1, 0.01)
        self.bounds = [(0, 1.), (0.01, 15), (0.01, 1.), (-1, 1), (0, 1.)]

    def calibrate(self, seed: int = None) -> Tuple[ql.HestonProcess, ql.HestonModel]:
        """Calibrate the Heston model."""
        process, model, engine = self._setup_model()
        helpers, grid_data = self._setup_helpers(engine)
        cost_function = self._cost_function_generator(model, helpers, norm=True)
        differential_evolution(cost_function, self.bounds, seed=seed, maxiter=100)
        print('Calibrated Heston parameters:', model.params())

        return process, model

    def _setup_model(self, initial_conditions: Tuple[float, ...] = None) -> Tuple[ql.HestonProcess,
                                                                                  ql.HestonModel,
                                                                                  ql.AnalyticHestonEngine]:
        if initial_conditions:
            theta, kappa, sigma, rho, v0 = initial_conditions
        else:
            theta, kappa, sigma, rho, v0 = self.initial_conditions

        process = ql.HestonProcess(self.market_data.yield_ts, self.market_data.dividend_ts,
                                   ql.QuoteHandle(ql.SimpleQuote(self.market_data.spot)),
                                   v0, kappa, theta, sigma, rho)
        model = ql.HestonModel(process)
        engine = ql.AnalyticHestonEngine(model)

        return process, model, engine

    def _setup_helpers(self, engine: ql.AnalyticHestonEngine) -> Tuple[List[ql.HestonModelHelper],
                                                                       List[Tuple[ql.Date, float]]]:
        helpers = []
        grid_data = []
        for i, date in enumerate(self.market_data.expiration_dates):
            for j, strike in enumerate(self.market_data.strikes):
                t = (date - self.reference_date)
                p = ql.Period(t, ql.Days)
                vols = self.market_data.data[i][j]
                helper = ql.HestonModelHelper(
                    p, self.calendar, self.market_data.spot, strike,
                    ql.QuoteHandle(ql.SimpleQuote(vols)),
                    self.market_data.yield_ts, self.market_data.dividend_ts)
                helper.setPricingEngine(engine)
                helpers.append(helper)
                grid_data.append((date, strike))

        return helpers, grid_data

    @staticmethod
    def _cost_function_generator(model: ql.HestonModel, helpers: List[ql.HestonModelHelper], norm=False):
        def cost_function(params):
            params_ = ql.Array(list(params))
            model.setParams(params_)
            error = [h.calibrationError() for h in helpers]
            if norm:
                return np.sqrt(np.sum(np.abs(error)))
            else:
                return error

        return cost_function

    @staticmethod
    def generate_paths(dates,
                       day_counter: ql.DayCounter,
                       process: ql.HestonProcess,
                       number_of_paths: int = 10000,
                       seed: int = None) -> np.ndarray:
        """Generate underlying and volatility paths."""
        dimension = process.factors()
        times = np.array([day_counter.yearFraction(dates[0], d) for d in dates])
        time_step = times.shape[0] - 1
        uniform_random_generator = ql.UniformRandomGenerator(seed=seed)
        sequence_generator = ql.UniformRandomSequenceGenerator(dimension * time_step, uniform_random_generator)
        gaussian_sequence_generator = ql.GaussianRandomSequenceGenerator(sequence_generator)
        paths_generator = ql.GaussianMultiPathGenerator(process, times, gaussian_sequence_generator)
        paths = np.zeros(shape=(number_of_paths, times.shape[0]))

        for i in range(number_of_paths):
            sample_path = paths_generator.next()
            values = sample_path.value()
            spot = values[0]
            # first argument refers to the underlying path, the second the volatility
            paths[i, :] = np.array(list(spot))

        # return array dimensions: [number of paths, number of items in t array]
        return paths



import QuantLib as ql
from typing import List


class MarketData(object):

    def __init__(self,
                 reference_date: ql.Date,
                 spot: float,
                 yield_ts: ql.YieldTermStructureHandle,
                 dividend_ts: ql.YieldTermStructureHandle,
                 expiration_dates: List[ql.Date] = None,
                 strikes: List[float] = None,
                 data: List[List[float]] = None,
                 black_scholes_volatility: float = None) -> None:
        self.reference_date = reference_date
        assert spot > 0
        self.spot = spot
        self.expiration_dates = expiration_dates
        self.strikes = strikes
        self.data = data  # volatility surface
        self.black_scholes_volatility = black_scholes_volatility
        self.yield_ts = yield_ts
        self.dividend_ts = dividend_ts



import QuantLib as ql
import numpy as np
from typing import List, Tuple
from exotx.data.marketdata import MarketData


class BlackScholesModel:
    """Class for the Black-Scholes model."""
    def __init__(self,
                 reference_date: ql.Date,
                 calendar: ql.Calendar,
                 market_data: MarketData) -> None:
        self.reference_date = reference_date
        self.calendar = calendar
        self.market_data = market_data

    def setup_model(self) -> Tuple[ql.BlackScholesMertonProcess, ql.AnalyticEuropeanEngine]:
        spot_handle = ql.QuoteHandle(ql.SimpleQuote(self.market_data.spot))
        flat_ts = self.market_data.yield_ts
        dividend_yield = self.market_data.dividend_ts
        flat_vol_ts = ql.BlackVolTermStructureHandle(
            ql.BlackConstantVol(self.reference_date,
                                self.calendar,
                                self.market_data.black_scholes_volatility,
                                ql.Actual365Fixed())
        )
        process = ql.BlackScholesMertonProcess(spot_handle, dividend_yield, flat_ts, flat_vol_ts)
        engine = ql.AnalyticEuropeanEngine(process)

        return process, engine

    @staticmethod
    def generate_paths(dates,
                       day_counter: ql.DayCounter,
                       process: ql.BlackScholesMertonProcess,
                       number_of_paths: int = 100000,
                       seed: int = None) -> np.ndarray:
        """Generate underlying and volatility paths."""
        dimension = process.factors()
        times = np.array([day_counter.yearFraction(dates[0], d) for d in dates])
        time_step = times.shape[0] - 1
        uniform_random_generator = ql.UniformRandomGenerator(seed=seed)
        sequence_generator = ql.UniformRandomSequenceGenerator(dimension * time_step, uniform_random_generator)
        gaussian_sequence_generator = ql.GaussianRandomSequenceGenerator(sequence_generator)
        paths_generator = ql.GaussianMultiPathGenerator(process, times, gaussian_sequence_generator)
        paths = np.zeros(shape=(number_of_paths, times.shape[0]))

        for i in range(number_of_paths):
            sample_path = paths_generator.next()
            values = sample_path.value()
            spot = values[0]
            # first argument refers to the underlying path, the second the volatility
            paths[i, :] = np.array(list(spot))

        return paths



import pytest
import QuantLib as ql
from exotx.data.marketdata import MarketData
from exotx.instruments.autocallable import Autocallable
from exotx.models.hestonmodel import HestonModel


# Arrange
@pytest.fixture
def my_reference_date():
    return ql.Date(6, 11, 2015)


@pytest.fixture
def my_market_data():
    day_count = ql.Actual365Fixed()
    reference_date = ql.Date(6, 11, 2015)
    spot = 100.0
    black_scholes_volatility = 0.2
    risk_free_rate = 0.01
    dividend_rate = 0.0
    yield_ts = ql.YieldTermStructureHandle(
        ql.FlatForward(reference_date, risk_free_rate, day_count))
    dividend_ts = ql.YieldTermStructureHandle(
        ql.FlatForward(reference_date, dividend_rate, day_count))

    return MarketData(reference_date=reference_date,
                      spot=spot,
                      yield_ts=yield_ts,
                      dividend_ts=dividend_ts,
                      black_scholes_volatility=black_scholes_volatility)


@pytest.fixture
def my_autocallable():
    notional = 100
    strike = 100
    autocall_barrier_level = 1.0
    coupon = 0.03
    coupon_barrier_level = 0.75
    protection_barrier_level = 0.75

    return Autocallable(notional, strike, autocall_barrier_level, coupon, coupon_barrier_level, protection_barrier_level)


def test_autocallable_black_scholes_price(my_autocallable: Autocallable,
                                          my_reference_date: ql.Date,
                                          my_market_data: MarketData):
    # Act
    seed = 125
    model = 'black-scholes'
    pv = my_autocallable.price(my_reference_date, my_market_data, model, seed)

    # Assert
    assert pv == pytest.approx(96.1609093840671, abs=1e-10)



"""Modules for working with pricing models."""

from exotx.models.blackscholesmodel import BlackScholesModel
from exotx.models.hestonmodel import HestonModel

__all__ = [
    'BlackScholesModel',
    'HestonModel'
]


"""Modules for working with exotic options.

Exotic options can refer to auto-callables, barrier options, etc."""

from exotx.instruments.autocallable import Autocallable

__all__ = [
    'Autocallable',
]


from setuptools import setup, find_packages

with open('README.md', 'r', encoding="utf-8", errors='ignore') as fh:
    long_description = fh.read()

version = {}
with open("exotx/_version.py", encoding="utf-8") as fp:
    exec(fp.read(), version)

setup(name='exotx',
      version=version['__version__'],
      description='Python library for pricing autocallables',
      author='Sebastien Eveno',
      author_email='sebastien.louis.eveno@gmail.com',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/SebastienEveno/exotx',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'numpy>=1.16.5',
          'pandas',
          'scipy',
          'matplotlib',
          'plotly>=4.12.0',
          'quantlib>=1.26'
      ],
      python_requires='>=3.6, <4',
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development',
        'Topic :: Office/Business :: Financial',
        'Topic :: Scientific/Engineering :: Information Analysis'
        ]
      )


import QuantLib as ql
import numpy as np
from typing import List
from datetime import datetime

from exotx.data.marketdata import MarketData
from exotx.models.blackscholesmodel import BlackScholesModel
from exotx.models.hestonmodel import HestonModel


class Autocallable:
    """Class for modeling an autocallable and pricing it."""
    def __init__(self,
                 notional: int,
                 # observation_dates: List[datetime],
                 # redemption_levels: List[float],
                 strike: float,
                 autocall_barrier_level: float,
                 annual_coupon_value: float,
                 coupon_barrier_level: float,
                 # coupon_barrier_levels: List[float],
                 protection_barrier_level: float,
                 # annual_coupon_value: float,
                 has_memory: bool = False) -> None:
        assert notional > 0, "Notional equal to zero or is negative!"
        self.notional = notional

        assert strike > 0, "strike equal to zero or is negative!"
        self.strike = strike

        assert autocall_barrier_level > 0, "Autocall barrier level equal to zero or is negative!"
        self.autocall_barrier_level = autocall_barrier_level

        assert annual_coupon_value > 0, "Annual coupon value equal to zero or is negative!"
        self.annual_coupon_value = annual_coupon_value

        # assert len(observation_dates) > 0, "No observation dates!"
        # self.observation_dates = list(set(observation_dates))
        #
        # assert len(redemption_levels) > 0, "No redemption levels!"
        # self.redemption_levels = redemption_levels

        # assert len(coupon_barrier_levels) > 0, "No coupon barrier levels!"
        # self.coupon_barrier_levels = coupon_barrier_levels

        assert coupon_barrier_level > 0, "Coupon barrier level equal to zero or is negative!"
        self.coupon_barrier_level = coupon_barrier_level

        self.protection_barrier_level = protection_barrier_level
        # self.annual_coupon_value = annual_coupon_value
        self.has_memory = has_memory

    @staticmethod
    def _get_underlying_paths(reference_date: ql.Date,
                              dates: np.ndarray,
                              calendar: ql.Calendar,
                              day_counter: ql.DayCounter,
                              market_data: MarketData,
                              model: str,
                              seed: int = None) -> np.ndarray:
        if model.lower() == 'black-scholes':
            black_scholes_model = BlackScholesModel(reference_date, calendar, market_data)
            process, model = black_scholes_model.setup_model()
            underlying_paths = black_scholes_model.generate_paths(dates, day_counter, process, seed=seed)[:, 1:]
        else:
            # defaults to Heston model
            # create and calibrate the heston model based on market data
            heston_model = HestonModel(reference_date, calendar, market_data)
            process, model = heston_model.calibrate(seed=seed)
            # generate paths for a given set of dates, exclude the current spot rate
            underlying_paths = heston_model.generate_paths(dates, day_counter, process, seed=seed)[:, 1:]

        return underlying_paths

    def price(self, reference_date: ql.Date, market_data: MarketData, model: str, seed: int = None):
        ql.Settings.instance().evaluationDate = reference_date
        convention = ql.ModifiedFollowing
        day_counter = ql.Actual365Fixed()
        calendar = ql.TARGET()

        # coupon schedule
        start_date = reference_date
        first_coupon_date = calendar.advance(start_date, ql.Period(6, ql.Months))
        last_coupon_date = calendar.advance(start_date, ql.Period(3, ql.Years))
        coupon_dates = np.array(list(ql.Schedule(first_coupon_date, last_coupon_date, ql.Period(ql.Semiannual),
                                                 calendar, convention, convention,
                                                 ql.DateGeneration.Forward, False)))
        # create past fixings into dictionary
        past_fixings = {}

        # immediate exit trigger for matured transaction
        if reference_date >= coupon_dates[-1]:
            return 0.0

        # immediate exit trigger for any past autocall event
        if reference_date >= coupon_dates[0]:
            if max(past_fixings.values()) >= (self.autocall_barrier_level * self.strike):
                return 0.0

        # create date array for path generator
        # combine valuation date and all the remaining coupon dates
        dates = np.hstack((np.array([reference_date]), coupon_dates[coupon_dates > reference_date]))

        # get underlying paths
        paths = self._get_underlying_paths(reference_date, dates, calendar, day_counter, market_data, model, seed)

        # identify the past coupon dates
        past_coupon_dates = coupon_dates[coupon_dates <= reference_date]

        # conditionally, merge given past fixings from a given dictionary and generated paths
        if past_coupon_dates.shape[0] > 0:
            past_fixings_array = np.array([past_fixings[past_date] for past_date in past_coupon_dates])
            past_fixings_array = np.tile(past_fixings_array, (paths.shape[0], 1))
            paths = np.hstack((past_fixings_array, paths))

        # result accumulator
        global_pv = []
        expiration_date = coupon_dates[-1]
        has_memory = int(self.has_memory)

        # loop through all simulated paths
        for path in paths:
            payoff_present_value = 0.0
            unpaid_coupons = 0  # counter of unpaid coupons
            has_auto_called = False

            # loop through set of coupon dates and index ratios
            for date, index in zip(coupon_dates, (path / self.strike)):
                # if autocall event has been triggered, immediate exit from this path
                if has_auto_called:
                    break

                payoff = 0.0
                # payoff calculation at expiration
                if date == expiration_date:
                    # index is greater or equal to coupon barrier level
                    # pay 100% redemption, plus coupon, plus conditionally all unpaid coupons
                    if index >= self.coupon_barrier_level:
                        payoff = self.notional * (1 + (self.annual_coupon_value / 2 * (1 + unpaid_coupons * has_memory)))
                    # index is greater or equal to protection barrier level and less than coupon barrier level
                    # pay 100% redemption, no coupon
                    if (index >= self.protection_barrier_level) & (index < self.coupon_barrier_level):
                        payoff = self.notional
                    # index is less than protection barrier
                    # pay redemption according to formula, no coupon
                    if index < self.protection_barrier_level:
                        payoff = self.notional * index
                # payoff calculation before expiration
                else:
                    # index is greater or equal to autocall barrier level
                    # autocall will happen before expiration
                    # pay 100% redemption, plus coupon, plus conditionally all unpaid coupons
                    if index >= self.autocall_barrier_level:
                        payoff = self.notional * (1 + (self.annual_coupon_value / 2 * (1 + unpaid_coupons * has_memory)))
                        has_auto_called = True
                    # index is greater or equal to coupon barrier level and less than autocall barrier level
                    # autocall will not happen
                    # pay coupon, plus conditionally all unpaid coupons
                    if (index >= self.coupon_barrier_level) & (index < self.autocall_barrier_level):
                        payoff = self.notional * (self.annual_coupon_value / 2 * (1 + unpaid_coupons * has_memory))
                        unpaid_coupons = 0
                    # index is less than coupon barrier level
                    # autocall will not happen
                    # no coupon payment, only accumulate unpaid coupons
                    if index < self.coupon_barrier_level:
                        payoff = 0.0
                        unpaid_coupons += 1

                # conditionally, calculate PV for period payoff, add PV to local accumulator
                if date > reference_date:
                    df = market_data.yield_ts.discount(date)
                    payoff_present_value += payoff * df

            # add path PV to global accumulator
            global_pv.append(payoff_present_value)

        return np.mean(np.array(global_pv))


